/*
** popen code written by Dave Bilitch
*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>

extern int	errno;

#define	FALSE		0
#define	TRUE		(!FALSE)

#define	MAX_PFILES	10

typedef struct _pfile
  {
    FILE	*pf_fs;
    int		pf_pid;
  } PFILE;

static PFILE	PFiles[MAX_PFILES];

/*
** popen
**
**	Local version of popen which allows for multiple simultaneous
**	popen calls.
*/

FILE	*popen(cmd, mode)

char	*cmd;
char	*mode;

{	/* popen --- Open pipe to/from command */

	FILE	*fs;
	int	pid;
	int	p[2];
	int	slot;

	if (*mode != 'w' && *mode != 'r')
	  {
	    errno = EINVAL;
	    return(NULL);
	  }

	/*
	** Find an empty slot in PFiles
	*/

	for (slot = 0; slot < MAX_PFILES; slot++)
	    if (PFiles[slot].pf_fs == NULL)
		break;

	if (slot >= MAX_PFILES)
	  {
	    errno = ENOSPC;
	    return(NULL);
	  }

	if (pipe(p) == -1)
	    return(NULL);

	fflush(stdout);
	pid = vfork();
	if (pid == -1)
	  {
	    close(p[0]);
	    close(p[1]);
	    return(NULL);
	  }

	if (pid == 0)
	  {
	    if (*mode == 'w')
	      {
		close(p[1]);
		close(0);
		dup(p[0]);
		close(p[0]);
	      }
	    else
	      {
		close(p[0]);
		close(1);
		dup(p[1]);
		close(p[1]);
	      }

	    execl("/bin/sh", "sh", "-c", cmd, NULL);
	    _exit(1);
	  }

	if (*mode == 'w')
	  {
	    close(p[0]);
	    fs = (FILE *) fdopen(p[1], mode);
	  }
	else
	  {
	    close(p[1]);
	    fs = (FILE *) fdopen(p[0], mode);
	  }

	PFiles[slot].pf_fs = fs;
	PFiles[slot].pf_pid = pid;

	return(fs);

}	/* popen */



pclose(fs)

FILE	*fs;

{	/* pclose --- Close the pipe to the specified program, wait on proc */

	int	pid;
	int	slot;
	int	i;

	for (slot = 0; slot < MAX_PFILES; slot++)
	    if (PFiles[slot].pf_fs == fs)
		break;

	if (slot >= MAX_PFILES)
	    return;

	fclose(fs);
	PFiles[slot].pf_fs = NULL;

	pid = WaitOn(PFiles[slot].pf_pid, NULL); 
	PFiles[slot].pf_pid = 0;

}	/* pclose */
